"""
http://stanford-clark.com/MQTT/
"""
import asyncio
import json
import random
from asyncio import StreamReader, StreamWriter
from typing import Optional

CONNECT = 1
CONNACK = 2
PUBLISH = 3
PUBACK = 4
SUBSCRIBE = 8
SUBACK = 9
PINGREQ = 12
PINGRESP = 13
DISCONNECT = 14


class MQTTMessage:
    type: int
    dup: bool
    qos: int
    retain: bool

    topic: str
    payload: bytes

    @property
    def text(self) -> str:
        return self.payload.decode()

    @property
    def json(self) -> dict:
        return json.loads(self.payload)

    def __str__(self):
        return f"{self.topic} {self.payload.decode()}"


class RawMessage:
    def __init__(self, raw=b""):
        self.pos = 0
        self.raw = raw

    @property
    def size(self):
        return len(self.raw)

    def read(self, length: int) -> bytes:
        self.pos += length
        return self.raw[self.pos - length : self.pos]

    def read_int(self, length: int) -> int:
        return int.from_bytes(self.read(length), "big")

    def read_str(self) -> str:
        slen = self.read_int(2)
        return self.read(slen).decode()

    def read_all(self) -> bytes:
        return self.read(self.size - self.pos)

    def write_int(self, value: int, length: int):
        self.raw += value.to_bytes(length, "big")

    def write_str(self, value: str):
        self.write_int(len(value), 2)
        self.raw += value.encode()

    def write_len(self):
        buf = b""
        var = len(self.raw)
        for _ in range(4):
            if var >= 128:
                var, b = divmod(var, 128)
                buf += (b | 128).to_bytes(1, "big")
            else:
                buf += var.to_bytes(1, "big")
                break
        self.raw = buf + self.raw

    def write_header(self, msg_type: int, qos=0, retain=False):
        self.write_len()
        header = (msg_type << 4) | (qos << 1) | int(retain)
        self.raw = header.to_bytes(1, "big") + self.raw

    @staticmethod
    def read_header(header: int) -> MQTTMessage:
        msg = MQTTMessage()
        msg.type = (header >> 4) & 0b1111
        msg.dup = bool((header >> 3) & 1)
        msg.qos = (header >> 1) & 0b11
        msg.retain = bool(header & 1)
        return msg

    @staticmethod
    def connect(keep_alive: int = 0):
        msg = RawMessage()
        msg.write_str("MQIsdp")  # protocol name
        msg.write_int(3, 1)  # protocol version
        msg.write_int(0, 1)  # flags
        msg.write_int(keep_alive, 2)  # keep alive
        cid = random.randint(1000, 9999)
        msg.write_str(f"hass-{cid}")  # client ID (should be unique)
        msg.write_header(CONNECT, qos=0)
        return msg.raw

    @staticmethod
    def subscribe(msg_id: int, *topics, qos=0):
        msg = RawMessage()
        msg.write_int(msg_id, 2)  # message ID
        for topic in topics:
            msg.write_str(topic)
            msg.write_int(qos, 1)  # requested QoS
        msg.write_header(SUBSCRIBE, qos=1)
        return msg.raw

    @staticmethod
    def publish(topic: str, payload: bytes, retain=False):
        msg = RawMessage()
        msg.write_str(topic)
        # skip msg_id for QoS 0
        msg.raw += payload
        msg.write_header(PUBLISH, qos=0, retain=retain)
        return msg.raw

    @staticmethod
    def ping():
        # adds zero length after header
        return (PINGREQ << 4).to_bytes(2, "little")

    @staticmethod
    def disconnect():
        # adds zero length after header
        return (DISCONNECT << 4).to_bytes(2, "little")


class MiniMQTT:
    msg_id: int = None
    reader: StreamReader = None
    writer: StreamWriter = None

    def __init__(self, keepalive=15, timeout=5):
        self.keepalive = keepalive
        self.timeout = timeout

    async def read_varlen(self) -> int:
        var = 0
        for i in range(4):
            b = await self.reader.read(1)
            var += (b[0] & 0x7F) << (7 * i)
            if (b[0] & 0x80) == 0:
                break
        return var

    async def _connect(self, host: str):
        self.reader, self.writer = await asyncio.open_connection(host, 1883)

        msg = RawMessage.connect()
        self.writer.write(msg)
        await self.writer.drain()

        self.msg_id = 0

        raw = await self.reader.readexactly(4)
        assert raw[0] == CONNACK << 4
        assert raw[1] == 2
        return raw[3] == 0

    async def connect(self, host: str):
        try:
            return await asyncio.wait_for(self._connect(host), self.timeout)
        except:
            return False

    async def disconnect(self):
        msg = RawMessage.disconnect()
        self.writer.write(msg)
        await self.writer.drain()

    async def subscribe(self, topic: str):
        self.msg_id += 1
        msg = RawMessage.subscribe(self.msg_id, topic)
        self.writer.write(msg)
        await self.writer.drain()

    async def publish(self, topic: str, payload, retain=False):
        if isinstance(payload, str):
            payload = payload.encode()
        elif isinstance(payload, dict):
            payload = json.dumps(payload, separators=(",", ":")).encode()

        # no response for QoS 0
        msg = RawMessage.publish(topic, payload, retain)
        self.writer.write(msg)
        await self.writer.drain()

    async def read(self) -> Optional[MQTTMessage]:
        raw = await self.reader.read(1)
        if raw == b"":
            # disconnected
            return None

        msg = RawMessage.read_header(raw[0])
        if msg.type == PUBLISH:
            varlen = await self.read_varlen()
            raw = await self.reader.readexactly(varlen)

            pr = RawMessage(raw)
            msg.topic = pr.read_str()

            if msg.qos > 0:
                _ = pr.read_int(2)  # msg ID
                raise NotImplementedError

            msg.payload = pr.read_all()

        elif msg.type == PINGRESP:
            await self.reader.readexactly(1)

        elif msg.type == SUBACK:
            # 1b header, 1b len, 2b msgID, 1b QOS
            varlen = await self.reader.readexactly(1)
            await self.reader.readexactly(varlen[0])

        else:
            raise NotImplementedError

        return msg

    async def close(self):
        if self.writer:
            self.writer.close()
            await self.writer.wait_closed()

    def __aiter__(self):
        return self

    async def __anext__(self) -> MQTTMessage:
        wait_pong = False

        while True:
            try:
                msg: MQTTMessage = await asyncio.wait_for(self.read(), self.keepalive)
                if msg is None:
                    raise StopAsyncIteration

                if msg.type == PUBLISH:
                    return msg

                if msg.type == PINGRESP:
                    wait_pong = False

            except asyncio.TimeoutError:
                if wait_pong:
                    # second ping without pong
                    raise StopAsyncIteration

                self.writer.write(RawMessage.ping())
                await self.writer.drain()

                wait_pong = True

            except Exception as e:
                raise e



from .const import (
    REGIONS,
    REGION_CANADA,
    REGION_EUROPE,
    REGION_USA,
    BRANDS,
    BRAND_KIA,
    BRAND_HYUNDAI,
)

from .KiaUvoApiImpl import KiaUvoApiImpl
from .KiaUvoApiCA import KiaUvoApiCA
from .KiaUvoApiEU import KiaUvoApiEU
from .KiaUvoAPIUSA import KiaUvoAPIUSA
from .HyundaiBlueLinkAPIUSA import HyundaiBlueLinkAPIUSA

DEFAULT_DISTANCE_UNIT_ARRAY = []


def get_default_distance_unit() -> int:
    return DEFAULT_DISTANCE_UNIT_ARRAY[0]


def get_implementation_by_region_brand(
    region: int,
    brand: int,
    username: str,
    password: str,
    use_email_with_geocode_api: bool = False,
    pin: str = "",
) -> KiaUvoApiImpl:  # pylint: disable=too-many-arguments
    if REGIONS[region] == REGION_CANADA:
        return KiaUvoApiCA(
            username, password, region, brand, use_email_with_geocode_api, pin
        )
    elif REGIONS[region] == REGION_EUROPE:
        return KiaUvoApiEU(
            username, password, region, brand, use_email_with_geocode_api, pin
        )
    elif REGIONS[region] == REGION_USA and BRANDS[brand] == BRAND_HYUNDAI:
        return HyundaiBlueLinkAPIUSA(
            username, password, region, brand, use_email_with_geocode_api, pin
        )
    elif REGIONS[region] == REGION_USA and BRANDS[brand] == BRAND_KIA:
        return KiaUvoAPIUSA(
            username, password, region, brand, use_email_with_geocode_api, pin
        )



from enum import Enum

from dateutil import tz

# Configuration Constants
DOMAIN: str = "kia_uvo"
CONF_PIN: str = "pin"
CONF_STORED_CREDENTIALS: str = "stored_credentials"
CONF_SCAN_INTERVAL: str = "scan_interval"
CONF_FORCE_SCAN_INTERVAL: str = "force_scan_interval"
CONF_NO_FORCE_SCAN_HOUR_START: str = "no_force_scan_hour_start"
CONF_NO_FORCE_SCAN_HOUR_FINISH: str = "no_force_scan_hour_finish"
CONF_ENABLE_GEOLOCATION_ENTITY: str = "enable_geolocation_entity"
CONF_USE_EMAIL_WITH_GEOCODE_API: str = "use_email_with_geocode_api"
CONF_BRAND: str = "brand"

# I have seen that many people can survive with receiving updates in every 30 minutes. Let's see how KIA will respond
DEFAULT_SCAN_INTERVAL: int = 30
# When vehicle is running/active, it will update its status regularly, so no need to force it. If it has not been running, we will force it every 240 minutes
DEFAULT_FORCE_SCAN_INTERVAL: int = 240
DEFAULT_NO_FORCE_SCAN_HOUR_START: int = 22
DEFAULT_NO_FORCE_SCAN_HOUR_FINISH: int = 6
DEFAULT_ENABLE_GEOLOCATION_ENTITY: bool = False
DEFAULT_USE_EMAIL_WITH_GEOCODE_API: bool = False
TIME_ZONE_EUROPE = tz.gettz("Europe/Berlin")

# Integration Setting Constants
PARALLEL_UPDATES: int = 1
CONFIG_FLOW_VERSION: int = 1
PLATFORMS = ["binary_sensor", "device_tracker", "sensor", "lock"]
TOPIC_UPDATE: str = f"{DOMAIN}_update_{0}"

# Home Assistant Data Storage Constants
DATA_VEHICLE_INSTANCE: str = "vehicle"  # Vehicle Instance
DATA_VEHICLE_LISTENER: str = (
    "vehicle_listener"  # Vehicle Topic Update Listener Unsubcribe Caller
)
DATA_CONFIG_UPDATE_LISTENER: str = (
    "config_update_listener"  # Config Options Update Listener Unsubcribe Caller
)

# action status delay constants
INITIAL_STATUS_DELAY_AFTER_COMMAND: int = 15
RECHECK_STATUS_DELAY_AFTER_COMMAND: int = 10
# Retry Specific Constants
START_FORCE_UPDATE_AFTER_COMMAND: int = 10  # Trigger first force update after a command
INTERVAL_FORCE_UPDATE_AFTER_COMMAND: int = 30  # Consecutive force update calls interval
COUNT_FORCE_UPDATE_AFTER_COMMAND: int = (
    5  # Number of force update calls after a command
)

# Sensor Specific Constants
NOT_APPLICABLE: str = "Not Applicable"
DATE_FORMAT: str = "%Y-%m-%d %H:%M:%S.%f"
DYNAMIC_DISTANCE_UNIT: str = "dynamic_distance_unit"
DISTANCE_UNITS = {1: "km", 3: "m"}
DYNAMIC_TEMP_UNIT: str = "dynamic_temp_unit"

REGION_EUROPE = "Europe"
REGION_CANADA = "Canada"
REGION_USA = "USA"
REGIONS = {1: REGION_EUROPE, 2: REGION_CANADA, 3: REGION_USA}
DEFAULT_REGION = 1
DEFAULT_PIN = ""

BRAND_KIA = "Kia"
BRAND_HYUNDAI = "Hyundai"
BRANDS = {1: BRAND_KIA, 2: BRAND_HYUNDAI}
DEFAULT_BRAND = 1

EU_TEMP_RANGE = [x * 0.5 for x in range(28, 60)]
CA_TEMP_RANGE = [x * 0.5 for x in range(32, 64)]
USA_TEMP_RANGE = range(62, 82)


class VEHICLE_ENGINE_TYPE(Enum):
    EV = 1
    PHEV = 2
    IC = 3


class VEHICLE_LOCK_ACTION(Enum):
    LOCK = "close"
    UNLOCK = "open"



import asyncio
import struct
import time

from .. import fhem, generic
from ..core import bluetoothle

DEFAULT_TIMEOUT = 1

HANDLE_RW_PASSWORD = 0x0048 - 1
HANDLE_R_WATERING = 0x0015 - 1
HANDLE_W_WATERING = 0x0013 - 1
HANDLE_R_BATTERY = 0x0039 - 1
HANDLE_R_TEMPERATURE = 0x003B - 1
HANDLE_R_MIN_TEMP = 0x003D - 1
HANDLE_R_MAX_TEMP = 0x003F - 1
HANDLE_R_FIRMWARE = 0x004E - 1
HANDLE_RW_DEVNAME = 0x0052 - 1
HANDLE_RW_ECO_PART1 = 0x0033 - 1
HANDLE_RW_ECO_PART2 = 0x0045 - 1
HANDLE_RW_TIME_OFFSET = 0x0035 - 1
HANDLE_R_MAC = 0x004A - 1
HANDLE_RW_INCREASEREDUCE = 0x0043 - 1
HANDLE_W_COMMITCODE = 0x0037 - 1
HANDLE_RW_TIMERS = [
    0x0017 - 1,
    0x0019 - 1,
    0x001B - 1,
    0x001D - 1,
    0x001F - 1,
    0x0021 - 1,
    0x0023 - 1,
    0x0031 - 1,
    0x0025 - 1,
    0x0027 - 1,
    0x0029 - 1,
    0x002B - 1,
    0x002D - 1,
    0x002F - 1,
]


class gfprobt(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._ble_lock = asyncio.Lock()
        self._conn = None
        set_conf = {
            "update": {"help": "Retrieve values from GF Pro BT"},
            "off": {"help": "Turn off watering"},
            "toggle": {"help": "Toggle on/off"},
            "on": {
                "args": ["onseconds"],
                "params": {
                    "onseconds": {"default": 0, "format": "int", "optional": True}
                },
                "help": (
                    "Turn on watering. You can use seconds as parameter to set the duration.<br>"
                    "Duration will be set directly on the device to have a safe switch off even if connectivity is broken"
                ),
            },
            "devicename": {"help": "Set the devicename of the irrigation control"},
            "adjust": {
                "args": ["percentage", "duration"],
                "params": {
                    "percentage": {"default": 0, "format": "int"},
                    "duration": {"default": 360, "format": " int", "optional": True},
                },
                "help": "Adjust the timings with parameters:<br>percentage<br>duration in hours",
            },
            "eco": {"help": "Activate/decactivate eco mode"},
        }
        self.set_set_config(set_conf)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 4:
            return "Usage: define irrigation_control fhempy gfprobt <MAC>"
        self._mac = args[3]
        self.hash["MAC"] = self._mac
        self._conn = bluetoothle.BluetoothLE(
            self.logger,
            self.hash,
            self._mac,
        )
        self.create_async_task(self.update_loop())

    async def Undefine(self, hash):
        if self._conn:
            await self._conn.disconnect()
        return await super().Undefine(self.hash)

    # write_password is running in another thread
    async def write_password(self):
        await self._conn.write_gatt_char(HANDLE_RW_PASSWORD, b"123456")

    async def set_update(self, hash, params):
        self.create_async_task(self.update_once())

    async def set_on(self, hash, params):
        onseconds = params["onseconds"]
        self.create_async_task(self._ble_set(self.async_on(onseconds)))

    async def set_off(self, hash, params):
        self.create_async_task(self._ble_set(self.async_off()))

    async def set_toggle(self, hash, params):
        self.create_async_task(self._ble_set(self.async_toggle()))

    async def set_adjust(self, hash, params):
        self.create_async_task(
            self._ble_set((self.async_adjust(params["percentage"], params["duration"])))
        )

    async def _ble_set(self, fct):
        async with self._ble_lock:
            await fct

    async def async_on(self, onseconds):
        await self.async_update_watering()
        if self._watering == 0:
            await self.async_toggle()

    async def async_off(self):
        await self.async_update_watering()
        if self._watering == 1:
            await self.async_toggle()

    async def async_toggle(self):
        await self._conn.write_gatt_char(HANDLE_W_WATERING, b"\x00")
        await self._conn.write_gatt_char(HANDLE_W_WATERING, b"\x01")

    async def async_adjust(self, percentage, duration):
        duration_sec = duration * 60 * 60

        duration_hex = duration_sec.to_bytes(6, "little")
        percentage_hex = percentage.to_bytes(2, "little")
        await self._conn.write_gatt_char(
            HANDLE_RW_INCREASEREDUCE, duration_hex + percentage_hex
        )

        await self.write_offset()
        await self.commit_code()

    async def update_loop(self):
        self._conn.register_connection_established_listener(self.write_password())
        await self._conn.connect()
        while True:
            try:
                await self.update_once()
            except asyncio.CancelledError:
                break
            except Exception:
                self.logger.exception("Failed to update readings")
            await asyncio.sleep(60)

    async def update_once(self):
        async with self._ble_lock:
            await self.async_update()
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "state", "on" if self._watering == 1 else "off"
        )
        await fhem.readingsBulkUpdateIfChanged(self.hash, "watering", self._watering)
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "adjust_hours", self._adjust_hours
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "adjust_percentage", self._adjust_perc
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "batteryVoltage", self._battery
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "temperature", self._temperature
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "min_temperature", self._min_temp
        )
        await fhem.readingsBulkUpdateIfChanged(
            self.hash, "max_temperature", self._max_temp
        )
        await fhem.readingsBulkUpdateIfChanged(self.hash, "firmware", self._firmware)
        await fhem.readingsBulkUpdateIfChanged(self.hash, "devname", self._devname)
        await fhem.readingsEndUpdate(self.hash, 1)

    async def async_update_watering(self):
        self._watering = struct.unpack(
            "<b", await self._conn.read_gatt_char(HANDLE_R_WATERING)
        )[0]

    async def async_update(self):
        await self.async_update_watering()
        self._battery = struct.unpack(
            "<h", await self._conn.read_gatt_char(HANDLE_R_BATTERY)
        )[0]
        self._temperature = struct.unpack(
            "<h", await self._conn.read_gatt_char(HANDLE_R_TEMPERATURE)
        )[0]
        self._min_temp = struct.unpack(
            "<h", await self._conn.read_gatt_char(HANDLE_R_MIN_TEMP)
        )[0]
        self._max_temp = struct.unpack(
            "<h", await self._conn.read_gatt_char(HANDLE_R_MAX_TEMP)
        )[0]
        self._firmware = await self._conn.read_gatt_char(HANDLE_R_FIRMWARE)
        self._firmware = str(self._firmware[1]) + "." + str(self._firmware[0])
        self._devname = await self._conn.read_gatt_char(HANDLE_RW_DEVNAME).decode(
            "utf-8"
        )
        self._eco = await self._conn.read_gatt_char(HANDLE_RW_ECO_PART1)
        self._eco += await self._conn.read_gatt_char(HANDLE_RW_ECO_PART2)
        self._timeoffset = struct.unpack(
            "<I", await self._conn.read_gatt_char(HANDLE_RW_TIME_OFFSET)
        )[0]
        self._devmac = str(await self._conn.read_gatt_char(HANDLE_R_MAC))
        self._increasereduce = await self._conn.read_gatt_char(HANDLE_RW_INCREASEREDUCE)
        self._adjust_hours = struct.unpack("<I", self._increasereduce[0:4])[0] / 3600
        self._adjust_perc = struct.unpack("<h", self._increasereduce[4:])[0]
        self._raw_timers = {}
        for handle_timer in HANDLE_RW_TIMERS:
            self._raw_timers[handle_timer] = await self._conn.read_gatt_char(
                handle_timer
            )
            self.logger.debug(self._raw_timers[handle_timer])

    async def write_offset(self):
        now = time.localtime()
        sec_since_mon = (
            now.tm_wday * 3600 * 24 + now.tm_hour * 3600 + now.tm_min * 60 + now.tm_sec
        )
        sec_since_mon = sec_since_mon.to_byte(4, "little")
        await self._conn.write_gatt_char(HANDLE_RW_TIME_OFFSET, sec_since_mon)

    async def commit_code(self):
        await self._conn.write_gatt_char(HANDLE_W_COMMITCODE, b"\x00")
        await self._conn.write_gatt_char(HANDLE_W_COMMITCODE, b"\x01")



import asyncio
import codecs

from .. import fhem, generic
from ..core import bluetoothle

# handle: 0x0002, char properties: 0x02, char value handle: 0x0003, uuid: 00002a00-0000-1000-8000-00805f9b34fb
# handle: 0x0004, char properties: 0x02, char value handle: 0x0005, uuid: 00002a01-0000-1000-8000-00805f9b34fb
# handle: 0x0006, char properties: 0x02, char value handle: 0x0007, uuid: 00002a04-0000-1000-8000-00805f9b34fb
# handle: 0x0008, char properties: 0x02, char value handle: 0x0009, uuid: 00002aa6-0000-1000-8000-00805f9b34fb
# handle: 0x000a, char properties: 0x02, char value handle: 0x000b, uuid: 00002ac9-0000-1000-8000-00805f9b34fb
# handle: 0x000d, char properties: 0x20, char value handle: 0x000e, uuid: 00002a05-0000-1000-8000-00805f9b34fb
# handle: 0x0011, char properties: 0x08, char value handle: 0x0012, uuid: f3300002-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x0013, char properties: 0x10, char value handle: 0x0014, uuid: f3300003-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x0016, char properties: 0x02, char value handle: 0x0017, uuid: f3300004-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x0018, char properties: 0x08, char value handle: 0x0019, uuid: f3300005-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x001a, char properties: 0x10, char value handle: 0x001b, uuid: f3300006-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x001d, char properties: 0x08, char value handle: 0x001e, uuid: f3300007-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x001f, char properties: 0x0a, char value handle: 0x0020, uuid: f3300008-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x0021, char properties: 0x0a, char value handle: 0x0022, uuid: f3300009-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x0023, char properties: 0x0a, char value handle: 0x0024, uuid: f3300010-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x0025, char properties: 0x02, char value handle: 0x0026, uuid: f3300011-0a29-b060-c591-bc4763b5c000
# handle: 0x0027, char properties: 0x08, char value handle: 0x0028, uuid: f3300020-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x002a, char properties: 0x02, char value handle: 0x002b, uuid: f3320010-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x002c, char properties: 0x02, char value handle: 0x002d, uuid: f3320011-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x002e, char properties: 0x02, char value handle: 0x002f, uuid: f3320015-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x0030, char properties: 0x02, char value handle: 0x0031, uuid: f3320016-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x0032, char properties: 0x02, char value handle: 0x0033, uuid: f3320017-f0a2-9b06-0c59-1bc4763b5c00
# handle: 0x0034, char properties: 0x0a, char value handle: 0x0035, uuid: f3320018-f0a2-9b06-0c59-1bc4763b5c00


class blue_connect(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._ble_lock = asyncio.Lock()
        self.task_update_loop = None
        self.ble_dev = None

        self.water_temp = "0"
        self.water_orp = "0"
        self.water_ph = "0"
        self.water_salt = "0"
        self.water_conductivity = "0"
        self.rssi = ""
        set_conf = {
            "measure": {"help": "Send signal to start measuring"},
            "restart": {"help": "Restart blue connect"},
        }
        self.set_set_config(set_conf)
        return

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) != 4:
            return "Usage: define my_blueconnect fhempy blue_connect MAC"
        self._mac = args[3]
        self.hash["MAC"] = self._mac

        self.ble_dev = bluetoothle.BluetoothLE(self.logger, self.hash, self._mac)
        self.ble_dev.register_notification_listener(self.received_notification)

        self.task_update_loop = self.create_async_task(self.update_loop())

    async def Undefine(self, hash):
        if self.task_update_loop:
            self.task_update_loop.cancel()
        if self.ble_dev:
            await self.ble_dev.disconnect()
        return await super().Undefine(self.hash)

    async def set_measure(self, hash, params):
        self.create_async_task(self.measure_once())

    async def set_restart(self, hash, params):
        self.create_async_task(self.restart())

    def received_notification(self, uuid, data):
        self.raw_measurement = codecs.encode(data, "hex")
        raw_temp = int(self.raw_measurement[4:6] + self.raw_measurement[2:4], 16)
        self.water_temp = float(raw_temp) / 100

        raw_ph = int(self.raw_measurement[8:10] + self.raw_measurement[6:8], 16)
        self.water_ph = round((float(0x0800) - float(raw_ph)) / 232 + 7, 2)

        raw_orp = int(self.raw_measurement[12:14] + self.raw_measurement[10:12], 16)
        self.water_orp = round(float(raw_orp) / 4)

        raw_salt = int(self.raw_measurement[16:18] + self.raw_measurement[14:16], 16)
        self.water_salt = round(float(raw_salt) / 25, 2)

        raw_conductivity = int(
            self.raw_measurement[20:22] + self.raw_measurement[18:20], 16
        )
        self.water_conductivity = round(float(raw_conductivity) / 0.4134)
        self.rssi = self.ble_dev.device.rssi

        self.create_async_task(self.update_readings())

    async def measure(self):
        # start measure
        await self.ble_dev.write_gatt_char(
            "F3300002-F0A2-9B06-0C59-1BC4763B5C00", b"\x01"
        )

    async def restart(self):
        # start measure
        await self.ble_dev.write_gatt_char(
            "f3300020-f0a2-9b06-0c59-1bc4763b5c00", b"\x31"
        )
        await self.ble_dev.disconnect()
        await asyncio.sleep(1)
        await self.ble_dev.connect()

    async def update_loop(self):
        if not self.ble_dev.is_connected:
            await self.ble_dev.connect()
        await self.ble_dev.connected.wait()

        while True:
            try:
                await self.measure_once()
            except asyncio.CancelledError:
                self.logger.info("Stopping update loop")
                return
            except Exception:
                self.logger.exception("Failed to update readings")
            await asyncio.sleep(7200)

    async def measure_once(self):
        async with self._ble_lock:
            self.water_temp = 0
            self.water_orp = 0
            self.water_ph = 0
            self.water_conductivity = 0
            self.water_salt = 0

            await self.measure()

    async def update_readings(self):
        await fhem.readingsBeginUpdate(self.hash)
        await fhem.readingsBulkUpdate(self.hash, "rssi", self.rssi)
        await fhem.readingsBulkUpdate(self.hash, "temperature", self.water_temp)
        await fhem.readingsBulkUpdate(self.hash, "ph", self.water_ph)
        await fhem.readingsBulkUpdate(self.hash, "orp", self.water_orp)
        await fhem.readingsBulkUpdate(self.hash, "raw", self.raw_measurement)
        await fhem.readingsBulkUpdate(
            self.hash, "conductivity", self.water_conductivity
        )
        await fhem.readingsBulkUpdate(self.hash, "salinity", self.water_salt)

        state = []
        if self.water_ph < 7.2:
            state.append(
                f"Ph: {self.water_ph} too low, diff {round(7.2-self.water_ph, 2)}"
            )
            await fhem.readingsBulkUpdate(self.hash, "ph_state", "low")
        elif self.water_ph > 7.6:
            state.append(
                f"Ph: {self.water_ph} too high, diff {round(7.6-self.water_ph, 2)}"
            )
            await fhem.readingsBulkUpdate(self.hash, "ph_state", "high")
        else:
            await fhem.readingsBulkUpdate(self.hash, "ph_state", "ok")

        if self.water_orp < 550:
            state.append(
                f"ORP: {self.water_orp} too low, diff {round(550-self.water_orp)}"
            )
            await fhem.readingsBulkUpdate(self.hash, "orp_state", "low")
        elif self.water_orp > 650:
            state.append(
                f"ORP: {self.water_orp} too high, diff {round(650-self.water_orp)}"
            )
            await fhem.readingsBulkUpdate(self.hash, "orp_state", "high")
        else:
            await fhem.readingsBulkUpdate(self.hash, "orp_state", "ok")

        if len(state) == 0:
            state = ["ok"]

        await fhem.readingsBulkUpdate(self.hash, "state", ", ".join(state))

        await fhem.readingsEndUpdate(self.hash, 1)



import logging

import requests
from dateutil import tz

from .const import (
    CA_TEMP_RANGE,
    EU_TEMP_RANGE,
    REGION_CANADA,
    REGION_EUROPE,
    REGION_USA,
    REGIONS,
    TIME_ZONE_EUROPE,
    USA_TEMP_RANGE,
)
from .Token import Token

_LOGGER = logging.getLogger(__name__)


class KiaUvoApiImpl:
    def __init__(
        self,
        username: str,
        password: str,
        region: int,
        brand: int,
        use_email_with_geocode_api: bool = False,
        pin: str = "",
    ):
        self.username = username
        self.password = password
        self.pin = pin
        self.use_email_with_geocode_api = use_email_with_geocode_api
        self.stamps = None
        self.region = region
        self.brand = brand
        self.last_action_tracked = False
        self.supports_soc_range = True

    def login(self) -> Token:
        pass

    def get_vehicles(self, token: Token):
        pass

    def get_cached_vehicle_status(self, token: Token):
        pass

    def check_last_action_status(self, token: Token):
        pass

    def get_geocoded_location(self, lat, lon):
        email_parameter = ""
        if self.use_email_with_geocode_api == True:
            email_parameter = "&email=" + self.username

        url = (
            "https://nominatim.openstreetmap.org/reverse?lat="
            + str(lat)
            + "&lon="
            + str(lon)
            + "&format=json&addressdetails=1&zoom=18"
            + email_parameter
        )
        response = requests.get(url)
        response = response.json()
        return response

    def update_vehicle_status(self, token: Token):
        pass

    def lock_action(self, token: Token, action):
        pass

    def start_climate(
        self, token: Token, set_temp, duration, defrost, climate, heating
    ):
        pass

    def stop_climate(self, token: Token):
        pass

    def start_charge(self, token: Token):
        pass

    def stop_charge(self, token: Token):
        pass

    def set_charge_limits(self, token: Token, ac_limit: int, dc_limit: int):
        pass

    def get_timezone_by_region(self):
        if REGIONS[self.region] == REGION_CANADA:
            return tz.UTC
        elif REGIONS[self.region] == REGION_EUROPE:
            return TIME_ZONE_EUROPE
        elif REGIONS[self.region] == REGION_USA:
            return tz.UTC

    def get_temperature_range_by_region(self):
        if REGIONS[self.region] == REGION_CANADA:
            return CA_TEMP_RANGE
        elif REGIONS[self.region] == REGION_EUROPE:
            return EU_TEMP_RANGE
        elif REGIONS[self.region] == REGION_USA:
            return USA_TEMP_RANGE



import ast
import asyncio
import functools
import json
import re

import tinytuya as tt

from .. import fhem, generic, utils
from . import mappings


class tuya(generic.FhemModule):
    def __init__(self, logger):
        super().__init__(logger)
        self._connected_device = None
        self.tuya_cloud = None
        self.tt_key = ""
        self.tt_secret = ""
        self.last_status = None
        self.create_device_list = []
        self.update_lock = asyncio.Lock()
        self.master_switch = ""
        self.update_dps_loop_task = None
        self.status_quick_loop_task = None

    # FHEM FUNCTION
    async def Define(self, hash, args, argsh):
        await super().Define(hash, args, argsh)
        if len(args) < 7:
            return (
                "Usage: define wifi_plug fhempy tuya"
                " setup <API_KEY> <API_SECRET> <DEVICE_ID> [<REGION>=eu]<br>"
                "OR if you want to define only one device with existing local key<br>"
                " <PRODUCTID> <DEVICE_ID> <IP> <LOCAL_KEY> "
                "[<VERSION>=3.3] [<API_KEY>] [<API_SECRET>]"
            )

        self.tt_type = args[3]
        if self.tt_type == "setup":
            self.tt_key = args[4]
            self.tt_secret = args[5]
            self.tt_did = args[6]
            if len(args) == 8:
                self.tt_region = args[7]
            else:
                self.tt_region = "eu"
            self.set_set_config({"scan_devices": {}})
            await fhem.readingsSingleUpdate(self.hash, "state", "ready", 1)
            # set internal
            hash["DEVICEID"] = 0
            hash["API_KEY"] = self.tt_key
            hash["API_SECRET"] = self.tt_secret
            hash["REGION"] = self.tt_region
            self.create_async_task(self.setup_cloud())
            return

        await fhem.readingsSingleUpdateIfChanged(self.hash, "online", "0", 1)
        self.tt_productid = self.tt_type
        self.tt_region = "eu"
        self.tt_did = args[4]
        self.tt_ip = args[5]
        self.tt_localkey = args[6]
        if len(args) >= 8:
            self.tt_version = float(args[7])
        else:
            self.tt_version = 3.3
        if len(args) == 10:
            self.tt_key = args[8]
            self.tt_secret = args[9]

        if self.tt_ip == "offline":
            await fhem.readingsSingleUpdate(
                self.hash, "state", "Change DEF and use IP instead of 'offline'", 1
            )
            return

        # set internal
        hash["DEVICEID"] = self.tt_did
        # set attributes
        self.attr_config = {
            "tuya_spec_functions": {"default": ""},
            "tuya_spec_status": {"default": ""},
        }
        self.set_attr_config(self.attr_config)
        # this is needed to set default values
        await utils.handle_define_attr(self.attr_config, self, hash)

        await self.setup_cloud()

        # create device
        self.create_async_task(self.create_device())

    async def setup_cloud(self):
        # create cloud device for cloud calls
        if self.tt_key and self.tt_secret:
            self.tuya_cloud = await utils.run_blocking(
                functools.partial(
                    tt.Cloud,
                    self.tt_region,
                    self.tt_key,
                    self.tt_secret,
                    self.tt_did,
                )
            )

    async def _create_mapping_dev(self):
        self.tuya_spec_functions = []
        self.tuya_spec_status = []
        schema = mappings.knownSchemas[self.tt_type]["schema"]
        for schema_part in schema:
            # rename to values
            schema_part["values"] = schema_part["property"]
            schema_part["type"] = schema_part["values"]["type"]
            schema_part["dp_id"] = schema_part["id"]
            schema_part["desc"] = ""
            if schema_part["type"] == "bool":
                schema_part["type"] = "Boolean"
            elif schema_part["type"] == "value":
                schema_part["type"] = "Integer"
            elif schema_part["type"] == "enum":
                schema_part["type"] = "Enum"
            elif schema_part["type"] == "string":
                schema_part["type"] = "String"
            else:
                continue
            self.tuya_spec_status.append(schema_part)
            if schema_part["mode"] == "rw":
                self.tuya_spec_functions.append(schema_part)

        await self._generate_set()

        await self.setup_connection()
        status = await self._connected_device.detect_available_dps()
        await self.update_readings(status, set_ready=True)

    async def set_attr_dp(self, hash):
        # check defined dp_Xs and their value
        # set set_conf based on the set dp_Xs
        dp_conf = {}
        for attr_name in self.__dict__:
            if attr_name.find("_attr_dp_") != -1:
                dp_id = re.search(r"\d+$", attr_name)[0]
                dp_conf[dp_id] = self.__dict__[attr_name]
                for fct in self.tuya_spec_functions:
                    if fct["code"] == dp_conf[dp_id]:
                        fct["id"] = int(dp_id)
                for status in self.tuya_spec_status:
                    if status["code"] == dp_conf[dp_id]:
                        status["id"] = int(dp_id)

        await self._generate_set()

    async def _generate_set(self):
        set_conf = {}

        # identify master switch
        for fct in self.tuya_spec_functions:
            if fct["code"] == "switch_1":
                self.master_switch = fct["code"]
            elif fct["code"] == "switch_led" and self.master_switch == "":
                self.master_switch = fct["code"]
            elif fct["code"][0:6] == "switch" and self.master_switch == "":
                self.master_switch = fct["code"]

        for fct in self.tuya_spec_functions:
            if "id" not in fct:
                # no id configured for this function
                continue
            if fct["type"] == "Boolean":
                if fct["code"] == self.master_switch:
                    set_conf["on"] = {
                        "help": fct["desc"],
                        "function_param": fct,
                        "function": "set_boolean",
                    }
                    set_conf["off"] = {
                        "help": fct["desc"],
                        "function_param": fct,
                        "function": "set_boolean",
                    }
                else:
                    set_conf[fct["code"]] = {
                        "options": "on,off",
                        "args": ["onoff"],
                        "help": fct["desc"],
                        "function_param": fct,
                        "function": "set_boolean",
                    }
            elif fct["type"] == "Enum":
                values = fct["values"]
                if "translation" in values:
                    options = list(values["translation"].values())
                else:
                    options = values["range"]
                set_conf[fct["code"]] = {
                    "options": ",".join(options),
                    "args": ["new_val"],
                    "help": fct["desc"],
                    "function_param": fct,
                    "function": "set_other_types",
                }
            elif fct["type"] == "Integer":
                spec = fct["values"]
                step = spec["step"] / (10 ** spec["scale"])
                if self.tt_productid == "IAYz2WK1th0cMLmL":
                    min = spec["min"] / 2
                    max = spec["max"] / 2
                else:
                    min = spec["min"] / (10 ** spec["scale"])
                    max = spec["max"] / (10 ** spec["scale"])
                slider = f"slider,{min},{step},{max}"
                set_conf[fct["code"]] = {
                    "options": slider,
                    "args": ["new_val"],
                    "params": {"new_val": {"format": "int"}},
                    "help": fct["desc"],
                    "function_param": fct,
                    "function": "set_integer",
                }
            elif fct["type"] == "String":
                set_conf[fct["code"]] = {
                    "args": ["new_val"],
                    "help": fct["desc"],
                    "function_param": fct,
                    "function": "set_other_types",
                }
            elif fct["type"] == "Json":
                set_conf[fct["code"]] = {
                    "args": ["new_val"],
                    "function_param": fct,
                    "function": "set_json",
                }
                if fct["code"] == "colour_data":
                    set_conf[fct["code"]]["function"] = "set_colour_data"
                    set_conf[fct["code"]]["options"] = "colorpicker,RGB"
                elif fct["code"] == "colour_data_v2":
                    set_conf[fct["code"]]["function"] = "set_colour_data_v2"
                    set_conf[fct["code"]]["options"] = "colorpicker,RGB"

        self.set_set_config(set_conf)

    async def set_create_device(self, hash, params):
        nameid = params["name"]
        for dev in self.create_device_list:
            if nameid == dev["name_esc"] + "_" + dev["device_id"]:
                await fhem.CommandDefine(
                    self.hash,
                    (
                        f"tuya_local_{dev['device_id']} fhempy tuya "
                        f"{dev['productid']} {dev['device_id']} {dev['ip']} "
                        f"{dev['local_key']} {dev['version']} "
                        f"{self.tt_key} {self.tt_secret}"
                    ),
                )
                await fhem.CommandAttr(
                    self.hash, f"tuya_local_{dev['device_id']} alias {dev['name']}"
                )

    async def set_boolean(self, hash, params):
        switch_id = params["function_param"]["id"]
        onoff = False
        if "onoff" in params:
            if params["onoff"] == "on":
                onoff = True
        else:
            if params["cmd"] == "on":
                onoff = True
        if self._connected_device:
            await self._connected_device.set_dp(onoff, switch_id)

    async def set_integer(self, hash, params):
        index = params["function_param"]["id"]
        if "translation" in params["function_param"]["values"]:
            for val in params["function_param"]["values"]["translation"].keys():
                if (
                    params["function_param"]["values"]["translation"][val]
                    == params["new_val"]
                ):
                    new_val = val
        else:
            if self.tt_productid == "IAYz2WK1th0cMLmL":
                if params["function_param"]["code"] == "temp_set":
                    params["new_val"] *= 0.2
            new_val = int(
                params["new_val"] * (10 ** params["function_param"]["values"]["scale"])
            )
        if self._connected_device:
            await self._connected_device.set_dp(new_val, index)

    async def set_other_types(self, hash, params):
        index = params["function_param"]["id"]
        if "translation" in params["function_param"]["values"]:
            for val in params["function_param"]["values"]["translation"].keys():
                if (
                    params["function_param"]["values"]["translation"][val]
                    == params["new_val"]
                ):
                    params["new_val"] = val
        new_val = params["new_val"]
        if self._connected_device:
            await self._connected_device.set_dp(new_val, index)

    async def set_colour_data(self, hash, params):
        index = params["function_param"]["id"]
        rgb = self.fhemrgb2rgb(params["new_val"])
        if "category" in self.info_dict and self.info_dict["category"] == "dj":
            hexvalue = tt.BulbDevice._rgb_to_hexvalue(rgb["r"], rgb["g"], rgb["b"], "A")
        else:
            hexvalue = tt.BulbDevice._rgb_to_hexvalue(rgb["r"], rgb["g"], rgb["b"], "B")
        await self.change_to_colour_mode()
        await self._connected_device.set_dp(hexvalue, index)

    async def set_colour_data_v2(self, hash, params):
        index = params["function_param"]["id"]
        rgb = self.fhemrgb2rgb(params["new_val"])
        hexvalue = tt.BulbDevice._rgb_to_hexvalue(rgb["r"], rgb["g"], rgb["b"], "B")
        await self.change_to_colour_mode()
        await self._connected_device.set_dp(hexvalue, index)

    async def change_to_colour_mode(self):
        if "work_mode" in self._conf_set:
            await self._connected_device.set_dp(
                "colour", self._conf_set["work_mode"]["function_param"]["id"]
            )

    def fhemrgb2rgb(self, rgb):
        red = int(rgb[0:2], base=16)
        green = int(rgb[2:4], base=16)
        blue = int(rgb[4:6], base=16)
        return {
            "r": red,
            "g": green,
            "b": blue,
        }

    # check if tuya_spec_functions and tuya_spec_status attr is set?
    async def check_tuya_attributes(self):
        self.tuya_spec_functions = await fhem.AttrVal(
            self.hash["NAME"], "tuya_spec_functions", "[]"
        )
        self.tuya_spec_status = await fhem.AttrVal(
            self.hash["NAME"], "tuya_spec_status", "[]"
        )
        self.tuya_spec_functions = ast.literal_eval(self.tuya_spec_functions)
        self.tuya_spec_status = ast.literal_eval(self.tuya_spec_status)

        self._convert_values_to_json()

    def _convert_values_to_json(self):
        for spec in self.tuya_spec_functions + self.tuya_spec_status:
            if spec["values"] == "":
                spec["values"] = "{}"
            if isinstance(spec["values"], str):
                spec["values"] = json.loads(spec["values"])

    # get functions/status from tuya
    async def get_tuya_dev_specification(self):
        # Get specification
        resp = await utils.run_blocking(
            functools.partial(self.tuya_cloud.getdps, self.tt_did)
        )
        self.logger.debug(f"getdps: {resp}")
        return resp["result"]

    async def get_tuya_dev_info(self):
        resp = await utils.run_blocking(
            functools.partial(self.tuya_cloud.getdevices, True)
        )
        if "Error" in resp:
            self.logger.error(f"getdevices: {resp}")
            return {}

        self.logger.debug(f"getdevices: {resp}")
        for dev in resp["result"]:
            if dev["id"] == self.tt_did:
                return dev
        return {}

    async def get_tuya_dev_description(self):
        # Get function description
        response_dict = await utils.run_blocking(
            functools.partial(self.tuya_cloud.getfunctions, self.tt_did)
        )
        self.logger.debug("getfunctions: f{response_dict}")
        fct_desc = response_dict["result"]["functions"]
        fct_code_desc = {}
        for fct in fct_desc:
            fct_code_desc[fct["code"]] = fct["desc"]
        return fct_code_desc

    async def async_status_updated(self, status):
        async with self.update_lock:
            await self.update_readings(status)

    def status_updated(self, status):
        self.create_async_task(self.async_status_updated(status))

    def disconnected(self):
        self.create_async_task(self.async_disconnected())

    async def _add_desc_to_spec(self, spec_fcts, desc):
        for spec in spec_fcts:
            if spec["code"] in desc:
                spec["desc"] = desc[spec["code"]]
        return spec_fcts

    async def retrieve_tuya_specs(self):
        # retrieve cloud codes
        spec = await self.get_tuya_dev_specification()
        self.tuya_spec_functions = spec["functions"]
        self.tuya_spec_status = spec["status"]
        self._convert_values_to_json()
        desc = await self.get_tuya_dev_description()
        self.tuya_spec_functions = await self._add_desc_to_spec(
            self.tuya_spec_functions, desc
        )
        await fhem.CommandAttr(
            self.hash,
            (
                f"{self.hash['NAME']} tuya_spec_functions "
                f"{str(self.tuya_spec_functions)}"
            ),
        )
        await fhem.CommandAttr(
            self.hash,
            f"{self.hash['NAME']} tuya_spec_status {str(self.tuya_spec_status)}",
        )

    async def _create_cloudmapping_dev(self):
        await fhem.readingsSingleUpdate(self.hash, "state", "Initializing...", 1)
        # check if attributes are set
        await self.check_tuya_attributes()
        if len(self.tuya_spec_functions) == 0 and len(self.tuya_spec_status) == 0:
            await self.retrieve_tuya_specs()
            info = await self.get_tuya_dev_info()
            await self.update_info_readings(info)
        else:
            self.info_dict = {
                "category": await fhem.ReadingsVal(self.hash["NAME"], "category", "")
            }

        await self.setup_connection()

        status = await self._connected_device.detect_available_dps()

        await self.prepare_attributes(status)
        await self.update_readings(status, set_ready=True)

    async def prepare_attributes(self, status):
        options = []
        dps = {}
        for opt in self.tuya_spec_status:
            options.append(opt["code"])
            dps[f"dp_{int(opt['dp_id']):02d}"] = ""

        for dp in list(status):
            dps[f"dp_{int(dp):02d}"] = ""

        attr_conf = {}
        for dp in list(dps):
            attr_conf[dp] = {
                "function": "set_attr_dp",
                "default": "",
                "options": ",".join(options),
            }
        self.attr_config.update(attr_conf)
        self.set_attr_config(self.attr_config)
        await utils.handle_define_attr(self.attr_config, self, self.hash)

        # if spec contains dp_id
        for spec in self.tuya_spec_status + self.tuya_spec_functions:
            if "dp_id" in spec:
                dp = spec["dp_id"]
                code = spec["code"]
                val = await fhem.AttrVal(self.hash["NAME"], f"dp_{int(dp):02d}", "")
                if val == "":
                    if f"dp_{int(dp):02d}" in self.attr_config:
                        await fhem.CommandAttr(
                            self.hash,
                            f"{self.hash['NAME']} dp_{int(dp):02d} {code}",
                        )
                    else:
                        self.logger.warning(
                            f"dp_{int(dp):02d} in spec but not found locally."
                            " This should be reported to TuYa"
                        )
        await self.set_attr_dp(self.hash)

    async def update_dps_loop(self):
        while True:
            await asyncio.sleep(5)
            # this is required to force update measurements (power, current, voltage)
            if self._connected_device:
                await self._connected_device.device.device.updatedps()

    async def status_quick_loop(self):
        while True:
            await asyncio.sleep(60)
            # this loop just ensures that all dps are updated every 60s
            if self._connected_device:
                await self._connected_device.device.device.status_quick()

    async def setup_connection(self):
        state_set = False
        connected = False
        while not connected:
            try:
                connect_fct = tt.connect(
                    self.tt_ip,
                    self.tt_did,
                    self.tt_localkey,
                    self.tt_version,
                    self,
                    timeout=15,
                )
                self._connected_device = await asyncio.wait_for(connect_fct, timeout=15)
                if self.update_dps_loop_task is None:
                    self.update_dps_loop_task = self.create_async_task(
                        self.update_dps_loop()
                    )
                if self.status_quick_loop_task is None:
                    self.status_quick_loop_task = self.create_async_task(
                        self.status_quick_loop()
                    )
                connected = True
            except Exception:
                if not state_set:
                    await fhem.readingsSingleUpdateIfChanged(
                        self.hash, "online", "0", 1
                    )
                    state_set = True
                    self.logger.exception("Failed to connect to device")
                # short sleep is required for passive devices
                await asyncio.sleep(1)

    async def create_device(self):
        try:
            if self.tt_key != "" and self.tt_secret != "":
                await self._create_cloudmapping_dev()
            elif self.tt_type in mappings.knownSchemas:
                await self._create_mapping_dev()
            else:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, "state", "Please use API_KEY and API_SECRET", 1
                )
        except Exception:
            self.logger.exception("Failed create_device")

    async def async_disconnected(self):
        await fhem.readingsSingleUpdate(self.hash, "online", "0", 1)
        self._connected_device = None
        await asyncio.sleep(5)
        await self.setup_connection()

    def convert(self, value, schema):
        if schema["type"] == "Integer":
            values = schema["values"]
            if self.tt_productid == "IAYz2WK1th0cMLmL":
                if schema["code"] == "cur_voltage":
                    value /= 0.2
                elif schema["code"] == "upper_temp":
                    value /= 10
                elif schema["code"] == "temp_set":
                    value /= 0.2
            elif self.tt_productid in ["wifvoilfrqeo6hvu", "37mnhia3pojleqfh"]:
                if schema["code"] == "cur_voltage":
                    value /= 10
            return value / (10 ** values["scale"])
        elif schema["type"] == "Boolean":
            if value is True:
                return "on"
            return "off"
        elif schema["type"] == "Enum":
            if "translation" in schema["values"]:
                if str(value) in schema["values"]["translation"]:
                    return schema["values"]["translation"][value]
        return value

    def alpha_to_dec(self, s):
        o = ord(s)
        if o >= ord("A") and o <= ord("Z"):
            return o - ord("A")
        if o >= ord("a") and o <= ord("z"):
            return o - ord("a") + 26
        if o >= ord("0") and o <= ord("9"):
            return o - ord("0") + 26 + 26
        if o == ord("/"):
            return 63
        return o

    def string_to_hexarr(self, s):
        resarr = []
        alphaarr = []
        for x in s:
            alphaarr.append(self.alpha_to_dec(x))
        x = 0
        while x < len(alphaarr):
            resarr.append(alphaarr[x] >> 2)
            resarr.append((alphaarr[x] & 0x03) << 2 | ((alphaarr[x + 1] & 0x30) >> 4))
            resarr.append(alphaarr[x + 1] & 0x0F)
            x += 2
        return resarr

    def convert_json(self, value, schema):
        flat_json = {}
        if "category" in self.info_dict and self.info_dict["category"] == "zndb":
            if schema["dp_id"] == 6:
                hexarr = self.string_to_hexarr(value)
                if len(hexarr) > 15:
                    flat_json["voltage"] = (
                        hexarr[1] << 8 | hexarr[2] << 4 | hexarr[3]
                    ) / 10
                    flat_json["current"] = (
                        hexarr[7] << 8 | hexarr[8] << 4 | hexarr[9]
                    ) / 1000
                    flat_json["power"] = hexarr[13] << 8 | hexarr[14] << 4 | hexarr[15]
        return flat_json

    async def update_info_readings(self, info_dict):
        self.info_dict = info_dict
        await fhem.readingsBeginUpdate(self.hash)
        try:
            for reading in info_dict:
                if reading == "status":
                    continue
                await fhem.readingsBulkUpdateIfChanged(
                    self.hash, reading, info_dict[reading]
                )
        except Exception:
            self.logger.exception("Failed to update info readings")
        await fhem.readingsEndUpdate(self.hash, 1)

    async def update_readings(self, status, set_ready=False):
        if "dps" in status:
            status = status["dps"]
        state_set = not set_ready
        await fhem.readingsBeginUpdate(self.hash)
        try:
            stateused = False
            for dp in status:
                found = False
                for st in self.tuya_spec_status:
                    if "dp_id" in st and st["dp_id"] == int(dp):
                        found = True
                        reading = st["code"]
                        if st["code"] == self.master_switch:
                            reading = "state"
                            stateused = True
                        self.logger.debug(
                            f"handle type {st['type']} for dp_id "
                            f"{st['dp_id']} with value {status[dp]}"
                        )
                        if st["type"] == "Json":
                            if st["code"] in ["colour_data", "colour_data_v2"]:
                                await self.update_readings_colour(
                                    st["code"], status[dp]
                                )
                            else:
                                flat_json = self.convert_json(status[dp], st)
                                for name in flat_json:
                                    await fhem.readingsBulkUpdateIfChanged(
                                        self.hash,
                                        reading + "_" + name,
                                        flat_json[name],
                                    )
                        else:
                            if reading == "state":
                                state_set = True
                            await fhem.readingsBulkUpdateIfChanged(
                                self.hash,
                                reading,
                                self.convert(status[dp], st),
                            )
                        break

                if not found:
                    await fhem.readingsBulkUpdateIfChanged(
                        self.hash,
                        f"dp_{int(dp):02d}",
                        status[dp],
                    )
            if not stateused:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "online", "1")
            if not state_set:
                await fhem.readingsBulkUpdateIfChanged(self.hash, "state", "ready")
        except Exception:
            self.logger.exception("Failed to update readings")
        await fhem.readingsEndUpdate(self.hash, 1)

    async def Undefine(self, hash):
        if self._connected_device:
            await self._connected_device.close()
            self._connected_device = None
        await super().Undefine(hash)

    # The following code is only for setup/scan process via tuya cloud
    async def set_scan_devices(self, hash, params):
        self.create_async_task(self._scan_devices())

    async def _scan_devices(self):
        json_data = await utils.run_blocking(
            functools.partial(self.tuya_cloud.getdevices, verbose=True)
        )
        if "Error" in json_data:
            self.logger.error(f"getdevices: {json_data}")
            await fhem.readingsSingleUpdate(self.hash, "state", f"{json_data}", 1)
            return

        # Filter to only Name, ID and Key, product_id, icon
        tuyadevices = []
        for i in json_data["result"]:
            item = {}
            item["name"] = i["name"].strip()
            item["id"] = i["id"]
            item["ip"] = i["ip"]
            item["sub"] = i["sub"]
            item["key"] = i["local_key"]
            item["product_id"] = i["product_id"]
            item["icon"] = i["icon"]
            await fhem.readingsBeginUpdate(self.hash)
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_name", item["name"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_id", item["id"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_localkey", item["key"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash, item["id"] + "_productid", item["product_id"]
            )
            await fhem.readingsBulkUpdateIfChanged(
                self.hash,
                item["id"] + "_icon",
                "https://images.tuyaeu.com/" + item["icon"],
            )
            await fhem.readingsEndUpdate(self.hash, 1)
            tuyadevices.append(item)

        # Display device list
        self.logger.debug("Device Listing")
        output = json.dumps(tuyadevices, indent=4)  # sort_keys=True)
        self.logger.debug(output)
        await fhem.readingsSingleUpdate(
            self.hash,
            "state",
            f"found {len(tuyadevices)} devices, start local scan...",
            1,
        )

        # scan local devices to get IP
        self.logger.debug("Scan local devices...")
        devices = await utils.run_blocking(
            functools.partial(tt.deviceScan, False, 20, False, False)
        )

        def getIP(d, gwid):
            for ip in d:
                if gwid == d[ip]["gwId"]:
                    return (ip, d[ip]["version"])
            return (0, 0)

        self.logger.debug("Polling local devices...")
        count_found = 0
        self.create_device_list = []
        for i in tuyadevices:
            name = i["name"]
            id = i["id"]
            (ip, ver) = getIP(devices, i["id"])
            local_key = i["key"]
            productid = i["product_id"]

            # do not create device if no IP was discovered
            if ver == 0 and i["sub"] is True:
                await fhem.readingsSingleUpdateIfChanged(
                    self.hash, f"{id}_ip", "offline", 1
                )
                continue

            if ver == 0:
                # device not discovered
                ip = "offline"
                ver = "3.3"
            else:
                count_found += 1

            await fhem.readingsSingleUpdateIfChanged(self.hash, f"{id}_ip", ip, 1)
            await fhem.readingsSingleUpdateIfChanged(
                self.hash, f"{id}_version", str(ver), 1
            )

            # create FHEM device
            self.create_device_list.append(
                {
                    "name": name,
                    "name_esc": name.replace(" ", "_"),
                    "device_id": id,
                    "productid": productid,
                    "ip": ip,
                    "local_key": local_key,
                    "version": ver,
                }
            )

        # devices which can be created from setup
        set_conf = {"scan_devices": {}}
        set_conf["create_device"] = {"args": ["name"], "options": ""}
        options = []
        for dev in self.create_device_list:
            options.append(dev["name_esc"] + "_" + dev["device_id"])
        set_conf["create_device"]["options"] = ",".join(options)
        self.set_set_config(set_conf)

        await fhem.readingsSingleUpdate(
            self.hash,
            "state",
            f"{count_found} devices found localy",
            1,
        )

    async def update_readings_colour(self, code, hexcolour):
        if code == "colour_data" and self.info_dict["category"] == "dj":
            # only category dj (light) has old colour_data
            (red, green, blue) = tt.BulbDevice._hexvalue_to_rgb(hexcolour, "A")
        else:
            (red, green, blue) = tt.BulbDevice._hexvalue_to_rgb(hexcolour, "B")

        rgb_hex = f"{red:02x}{green:02x}{blue:02x}"
        await fhem.readingsBulkUpdate(
            self.hash,
            code,
            rgb_hex,
        )
